/**
 * This is an example implementation of a stack-datastructure.
 * Copyright (c): Sascha Wille, inf6201@fh-wedel.de
 *
 * You may study, modify and distribute this source code
 * FOR NON-COMMERCIAL PURPOSES ONLY.
 * This copyright message has to remain unchanged.
 *
 * Note that this document is provided 'as is',
 * WITHOUT WARRANTY of any kind either expressed or implied.
 */
#include <python3.0/Python.h>

/* Forward-Declaration */
static PyTypeObject Stack_Type;

/**
 * StackObject datastructure
 *
 * PyObject_HEAD: common information of any Python-Object
 * maxSize      : stack-size (max number of elements on the stack)
 * len          : current number of elements of the stack
 * e            : array of stack-Elements
 */
typedef struct {
    PyObject_HEAD /* contains ref-count and typeinformation*/
    unsigned int maxSize;
    unsigned int len;
    PyObject ** e;
} StackObject;

/* StackError Exception object */
static PyObject *StackError;

/**
 * Sets StackError with given message
 *
 * @param[in] msg, String to be passed as exception-message
 * 
 * @return always NULL (error-indicator)
 */
static PyObject*
throw(char * msg) {
    PyErr_SetString(StackError, msg);
    return NULL;
}

/**
 * Returns the current size of the stack
 * 
 * @return number of elements on the stack
 */
static PyObject *
Stack_getSize(StackObject *self) {
    PyObject *s = Py_BuildValue("i", self->len);
    if (s == NULL) {
        return NULL;
    }
    return s;
}

/**
 * Returns the element on top of the stack.
 * NOTE! The element still remains on the stack
 *
 * @return Object from the top of the stack
 */
static PyObject *
Stack_top(StackObject *self) {
    PyObject *res = NULL;
    if (!self->len)
        return throw("Stack contains no elements.");

    res = self->e[self->len - 1];

    Py_INCREF(res);
    return res;
}

/**
 * Returns AND REMOVES the element on top of the stack.
 *
 * @return Object from the top of the stack
 */
static PyObject *
Stack_pop(StackObject *self) {
    PyObject *res = Stack_top(self);
    if (res) {
        --self->len;
        Py_DECREF(res);
    }

    /* No Py_INCREF, Stack_top() returns an owned reference */
    return res;
}

/**
 * Pushes ONE single element on top of the stack.
 *
 * @param[in] args, every possible python object (e.g. a sequence, list, ...)
 *
 * @return returns the new stack (owned reference)
 */
static PyObject *
Stack_push(StackObject *self, PyObject *args) {
    PyObject *e = NULL;

    /* Exception-Handling */
    if (self->len == self->maxSize)
        return throw("Stack overflow. Pop some elements or create a new stack.");

    if (!PyArg_ParseTuple(args, "O", &e)) {
        return NULL;
    }
    Py_INCREF(e);
    (self->e)[self->len] = e;
    ++self->len;

    Py_RETURN_NONE;
}

/**
 * Adds a sequence of objects to the stack (element-wise)
 *
 * @param[in,out] self, stack-objects to modify
 * @param[in] seq, PySequence of PyObjects to add
 *
 * @return errorcode, 0 -> ok, -1 -> error occured
 */
static int
Stack_addSequence(StackObject * self, PyObject *seq) {
    int err = 0;
    PyObject *tmpStack = NULL;

    /* push sequence elements on the stack */
    if (PySequence_Check(seq)) {
        int len = PySequence_Length(seq);
        int i;
        if (len > self->maxSize) {
            throw("Sequence too long for stacksize");
            err = -1;
        } else {
            for (i = 0; i < len; ++i) {
                PyObject* item = PySequence_GetItem(seq, i);
                if (item) {
                    tmpStack = Stack_push(self, Py_BuildValue("(O)", item));
                }
                Py_XDECREF(item);
                Py_XDECREF(tmpStack);
            }
        }
    } else {
        throw("sequence expected but not given");
        err = -1;
    }

    return err;
}

/**
 * determines the length of a sequence or just parses the integerobject
 *
 * @param[in] arg, Sequence or Long
 * @param[out] int, length of sequence or numeric representation of Long-Object
 *
 * @return 1 if params are correct, 0 otherwise
 */
static
unsigned int retrieveStackSize(PyObject *arg, int *stackSize) {
    unsigned int res = 1;
    if (PyLong_Check(arg)) {
        *stackSize = PyLong_AsSsize_t(arg);
    } else if (PySequence_Check(arg)) {
        *stackSize = PySequence_Length(arg);
    } else
        res = 0;

    return res;
}

/**
 * Allocation part of object-construction.
 * Allocates memory for the object an the internal object-array
 *
 * @param[in] type, Type of the object to create (maybe a subtype of Stack)
 * @param[in] args, constructor-call arguments.
 *
 * @return a new empty! StackObject
 */
PyObject *tp_alloc(PyTypeObject *type, Py_ssize_t nitems) {
    StackObject *self = NULL;
    PyObject ** elements = NULL;

    int i = 0;
    int err = 0;

    /* allocate enough memory and set specific type */
    self = (StackObject *) type->tp_alloc(type, 0);
    elements = malloc(nitems * type->tp_itemsize);
    err = (elements == NULL);

    for (i = 0; i < nitems && !err; ++i) {
        elements[i] = malloc(type->tp_itemsize);
        err = (elements == NULL);
        ++i;
    }
    self->e = elements;
    self->maxSize = nitems; /* has to be done here, because args are already parsed */

    if (err) {
        Py_XDECREF(self);
        /* pass the exception to the interpreter */
        PyErr_NoMemory();
        return NULL;
    }
    

    return (PyObject *) self;
}

/**
 * Allocation part of object-construction
 *
 * @param[in] type, Type of the object to create (maybe a subtype of Stack)
 * @param[in] args, constructor-call arguments.
 *
 * @return a new empty! StackObject
 */
static PyObject *
Stack_new(PyTypeObject *type, PyObject *args) {
    int stacksize = 0;
    PyObject * arg = NULL;

    /* PyArg_ParseTuple already sets proper Exception */
    if (!PyArg_ParseTuple(args, "O", &arg))
        return NULL;
    if (!retrieveStackSize(arg, &stacksize))
        return NULL;
#ifdef DEBUG
    PySys_WriteStdout("construction (alloc-part): new stack(%d) created.\n", stacksize);
#endif
    return tp_alloc(type, stacksize);

}

/**
 * Stack-Constructor.
 * In Python it would be:
 * class cStack:
 *    def __init__(stacksize, seq=None):
 *      ...
 *
 * @param[in,out] self, new empty StackObject created by Stack_new
 * @param[in] args, constructor-call arguments
 *
 * @return 0 if stack was created properly
 */
static int
Stack_init(StackObject *self, PyObject *args) {
    PyObject *arg = NULL;
    int err = 0;

    /* Init */
    self->len = 0;

    /* Add Elements if given */
    if (!PyArg_ParseTuple(args, "O", &arg))
        err = -1;
    if (!err && !PyLong_Check(arg))
        err = Stack_addSequence(self, arg);
#ifdef DEBUG
    if (!err) {
        PySys_WriteStdout("construction (init-part): done\n");
    }
#endif
    return err;
}

/**
 * traverse the stack by using the visit-function
 * CYCLIC GARBAGE COLLECTION
 *
 * @param[in] self, stack-object
 * @param[in] self, visit-function
 * @param[in] self, arguments of visit function
 *
 * @return result of vistit function
 */
static int
Stack_traverse(StackObject *self, visitproc visit, void *arg) {
    unsigned int i;
    int res = 0;

    for (i = 0; (i < self->len) && !res; ++i) {
        if (self->e[i]) {
            res = visit(self->e[i], arg);
        }
    }

    return res;
}

/**
 * Decref's all elements of the stack, because of garbage collection
 *
 * @param[in,out] self, stack to compute
 */
static int
Stack_clear(StackObject* self) {
    unsigned int i;
    PyObject *tmp = NULL;
#ifdef DEBUG
    PySys_WriteStderr("Clear\n");
#endif
    /* Dealloc stack */
    for (i = 0; i < self->len; ++i) {
        if (self->e[i]) {
            tmp = self->e[i];
            self->e[i] = NULL;
            Py_DECREF(tmp);
        }
    }
    return 0;
}

/**
 * Deallocation-Function for a StackObejct.
 * Deallocates all reserved memory and decrements the referencecount for
 * the elements.
 *
 * @param[in,out] self, the StackObject
 */
static void
Stack_dealloc(StackObject *self) {
#ifdef DEBUG
    PySys_WriteStderr("%s", "Starting dealloc\n");
#endif
    Stack_clear(self);
    Py_TYPE(self)->tp_free((PyObject*) self);
#ifdef DEBUG
    PySys_WriteStderr("%s", "stack freed\n");
#endif
}

/**
 * Factory-method for empty Stack with a size of 5
 *
 * @return empty Stack with a size of 5
 */
static PyObject * Stack_mkEmpty5(void) {
    PyObject *res = NULL;
    PyObject *len = Py_BuildValue("(i)", 5);
    res = Stack_new(&Stack_Type, len);
    if (res) {
        Stack_init((StackObject *) res, len);
    }
    Py_DECREF(len);
    return res;
}

/* Bind python-methodenames to a C-function (class-methods) */
static PyMethodDef Stack_methods[] = {
    {"push", (PyCFunction) Stack_push, METH_VARARGS, PyDoc_STR("pushes a value to the top")},
    {"pop", (PyCFunction) Stack_pop, METH_NOARGS, PyDoc_STR("returns AND removes the top element")},
    {"top", (PyCFunction) Stack_top, METH_NOARGS, PyDoc_STR("returns the top element")},
    {"getSize", (PyCFunction) Stack_getSize, METH_NOARGS, PyDoc_STR("get stack size")},
    {NULL, NULL} /* sentinel */
};

/* Bind python-methodenames to C-function (module-methods) */
static PyMethodDef Stack_factory_methods[] = {
    {"mkEmpty5", (PyCFunction) Stack_mkEmpty5, METH_NOARGS, PyDoc_STR("creates empty stack with 5 elements")},
    {NULL, NULL} /* sentinel */
};


/* Documentationstring of this module -> sets globale "string"*/
PyDoc_STRVAR(module_doc,
        "This is a module of a stack implementation in C");

/* Module structure */
static struct PyModuleDef Stackmodule = {
    PyModuleDef_HEAD_INIT, /* common module information */
    "cStackMod", /* module name */
    module_doc, /* documentantion */
    -1, /* additional memory allocation --> none*/
    Stack_factory_methods, /* availible functions */
    NULL, NULL, NULL, NULL /* dealloc operations */
};

/**
 * Typeinformation-Structure for the StackObeject
 */
static PyTypeObject Stack_Type = {
    PyVarObject_HEAD_INIT(NULL, 0)
    "cStack", /*tp_name*/
    sizeof (StackObject), /*tp_basicsize*/
    sizeof (PyObject*), /*tp_itemsize*/
    /* methods */
    (destructor) Stack_dealloc, /*tp_dealloc*/
    0, /*tp_print*/
    0, /*tp_getattr*/
    0, /* tp_setattr */
    0, /*tp_compare*/
    0, /*tp_repr*/
    0, /*tp_as_number*/
    0, /*tp_as_sequence*/
    0, /*tp_as_mapping*/
    0, /*tp_hash*/
    0, /*tp_call*/
    0, /*tp_str*/
    0, /*tp_getattro*/
    0, /*tp_setattro*/
    0, /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /* tp_flags */ /* allows us to subclass this type */
    "cStack Type", /*tp_doc*/
    (traverseproc) Stack_traverse, /*tp_traverse*/
    (inquiry) Stack_clear, /*tp_clear*/
    0, /*tp_richcompare*/
    0, /*tp_weaklistoffset*/
    0, /*tp_iter*/
    0, /*tp_iternext*/
    Stack_methods, /*tp_methods*/
    0, /*tp_members*/
    0, /*tp_getset*/
    0, /*tp_base*/
    0, /*tp_dict*/
    0, /*tp_descr_get*/
    0, /*tp_descr_set*/
    0, /*tp_dictoffset*/
    (initproc) Stack_init, /*tp_init*/
    0, /*tp_alloc*/
    (newfunc) Stack_new, /*tp_new*/
    0, /*tp_free*/
    0, /*tp_is_gc*/
};

/**
 * Initializes the module.
 *
 * @return initialized module or NULL in error cases
 */
PyMODINIT_FUNC
PyInit_cStackMod(void) {
    PyObject *m = NULL;

    /* inheritance from "object" */
    if (PyType_Ready(&Stack_Type) == 0) {
        m = PyModule_Create(&Stackmodule);
        if (m) {
            Py_INCREF(&Stack_Type);
            PyModule_AddObject(m, Stack_Type.tp_name, (PyObject*) & Stack_Type);
        }
    }

    /* adding stack exception to module m */
    if (!m)
        return NULL;
    if (StackError == NULL) {
        StackError = PyErr_NewException("stack.error", NULL, NULL); /* NULL = baseclass will be "exception" */
        if (StackError) {
            Py_INCREF(StackError);
            PyModule_AddObject(m, "stack_error", StackError);
        } else
            return NULL;
    }
    return m;
}
